@@ -10,6 +10,7 @@
 # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
 # ANY KIND, either express or implied. See the License for the specific
 # language governing permissions and limitations under the License.
+import base64
 import re
 import time
 import logging
@@ -24,18 +25,22 @@ import socket
 import cgi
 
 import dateutil.parser
-from dateutil.tz import tzlocal, tzutc
+from dateutil.tz import tzutc
 
 import botocore
 import botocore.awsrequest
 import botocore.httpsession
-from botocore.compat import json, quote, zip_longest, urlsplit, urlunsplit
-from botocore.compat import OrderedDict, six, urlparse
+from botocore.compat import (
+        json, quote, zip_longest, urlsplit, urlunsplit, OrderedDict,
+        six, urlparse, get_tzinfo_options, get_md5, MD5_AVAILABLE
+)
 from botocore.vendored.six.moves.urllib.request import getproxies, proxy_bypass
 from botocore.exceptions import (
     InvalidExpressionError, ConfigNotFound, InvalidDNSNameError, ClientError,
     MetadataRetrievalError, EndpointConnectionError, ReadTimeoutError,
-    ConnectionClosedError, ConnectTimeoutError,
+    ConnectionClosedError, ConnectTimeoutError, UnsupportedS3ArnError,
+    UnsupportedS3AccesspointConfigurationError, SSOTokenLoadError,
+    InvalidRegionError,
 )
 
 logger = logging.getLogger(__name__)
@@ -254,9 +259,16 @@ class _RetriesExceededError(Exception):
     pass
 
 
+class BadIMDSRequestError(Exception):
+    def __init__(self, request):
+        self.request = request
+
+
 class IMDSFetcher(object):
 
     _RETRIES_EXCEEDED_ERROR_CLS = _RetriesExceededError
+    _TOKEN_PATH = 'latest/api/token'
+    _TOKEN_TTL = '21600'
 
     def __init__(self, timeout=DEFAULT_METADATA_SERVICE_TIMEOUT,
                  num_attempts=1, base_url=METADATA_BASE_URL,
@@ -274,29 +286,56 @@ class IMDSFetcher(object):
             proxies=get_environ_proxies(self._base_url),
         )
 
-    def _get_request(self, url_path, retry_func):
+    def _fetch_metadata_token(self):
+        self._assert_enabled()
+        url = self._base_url + self._TOKEN_PATH
+        headers = {
+            'x-aws-ec2-metadata-token-ttl-seconds': self._TOKEN_TTL,
+        }
+        self._add_user_agent(headers)
+        request = botocore.awsrequest.AWSRequest(
+            method='PUT', url=url, headers=headers)
+        for i in range(self._num_attempts):
+            try:
+                response = self._session.send(request.prepare())
+                if response.status_code == 200:
+                    return response.text
+                elif response.status_code in (404, 403, 405):
+                    return None
+                elif response.status_code in (400,):
+                    raise BadIMDSRequestError(request)
+            except ReadTimeoutError:
+                return None
+            except RETRYABLE_HTTP_ERRORS as e:
+                logger.debug(
+                    "Caught retryable HTTP exception while making metadata "
+                    "service request to %s: %s", url, e, exc_info=True)
+        return None
+
+    def _get_request(self, url_path, retry_func, token=None):
         """Make a get request to the Instance Metadata Service.
 
         :type url_path: str
         :param url_path: The path component of the URL to make a get request.
-            This arg is appended to the base_url taht was provided in the
+            This arg is appended to the base_url that was provided in the
             initializer.
 
         :type retry_func: callable
         :param retry_func: A function that takes the response as an argument
              and determines if it needs to retry. By default empty and non
              200 OK responses are retried.
-         """
-        if self._disabled:
-            logger.debug("Access to EC2 metadata has been disabled.")
-            raise self._RETRIES_EXCEEDED_ERROR_CLS()
+
+        :type token: str
+        :param token: Metadata token to send along with GET requests to IMDS.
+        """
+        self._assert_enabled()
         if retry_func is None:
             retry_func = self._default_retry
         url = self._base_url + url_path
         headers = {}
-        if self._user_agent is not None:
-            headers['User-Agent'] = self._user_agent
-
+        if token is not None:
+            headers['x-aws-ec2-metadata-token'] = token
+        self._add_user_agent(headers)
         for i in range(self._num_attempts):
             try:
                 request = botocore.awsrequest.AWSRequest(
@@ -310,6 +349,15 @@ class IMDSFetcher(object):
                     "service request to %s: %s", url, e, exc_info=True)
         raise self._RETRIES_EXCEEDED_ERROR_CLS()
 
+    def _add_user_agent(self, headers):
+        if self._user_agent is not None:
+            headers['User-Agent'] = self._user_agent
+
+    def _assert_enabled(self):
+        if self._disabled:
+            logger.debug("Access to EC2 metadata has been disabled.")
+            raise self._RETRIES_EXCEEDED_ERROR_CLS()
+
     def _default_retry(self, response):
         return (
             self._is_non_ok_response(response) or
@@ -350,8 +398,9 @@ class InstanceMetadataFetcher(IMDSFetcher):
 
     def retrieve_iam_role_credentials(self):
         try:
-            role_name = self._get_iam_role()
-            credentials = self._get_credentials(role_name)
+            token = self._fetch_metadata_token()
+            role_name = self._get_iam_role(token)
+            credentials = self._get_credentials(role_name, token)
             if self._contains_all_credential_fields(credentials):
                 return {
                     'role_name': role_name,
@@ -377,18 +426,22 @@ class InstanceMetadataFetcher(IMDSFetcher):
             logger.debug("Max number of attempts exceeded (%s) when "
                          "attempting to retrieve data from metadata service.",
                          self._num_attempts)
+        except BadIMDSRequestError as e:
+            logger.debug("Bad IMDS request: %s", e.request)
         return {}
 
-    def _get_iam_role(self):
+    def _get_iam_role(self, token=None):
         return self._get_request(
             url_path=self._URL_PATH,
-            retry_func=self._needs_retry_for_role_name
+            retry_func=self._needs_retry_for_role_name,
+            token=token,
         ).text
 
-    def _get_credentials(self, role_name):
+    def _get_credentials(self, role_name, token=None):
         r = self._get_request(
             url_path=self._URL_PATH + role_name,
-            retry_func=self._needs_retry_for_credentials
+            retry_func=self._needs_retry_for_credentials,
+            token=token,
         )
         return json.loads(r.text)
 
@@ -541,6 +594,25 @@ def percent_encode(input_str, safe=SAFE_CHARS):
     return quote(input_str, safe=safe)
 
 
+def _parse_timestamp_with_tzinfo(value, tzinfo):
+    """Parse timestamp with pluggable tzinfo options."""
+    if isinstance(value, (int, float)):
+        # Possibly an epoch time.
+        return datetime.datetime.fromtimestamp(value, tzinfo())
+    else:
+        try:
+            return datetime.datetime.fromtimestamp(float(value), tzinfo())
+        except (TypeError, ValueError):
+            pass
+    try:
+        # In certain cases, a timestamp marked with GMT can be parsed into a
+        # different time zone, so here we provide a context which will
+        # enforce that GMT == UTC.
+        return dateutil.parser.parse(value, tzinfos={'GMT': tzutc()})
+    except (TypeError, ValueError) as e:
+        raise ValueError('Invalid timestamp "%s": %s' % (value, e))
+
+
 def parse_timestamp(value):
     """Parse a timestamp into a datetime object.
 
@@ -553,21 +625,14 @@ def parse_timestamp(value):
     This will return a ``datetime.datetime`` object.
 
     """
-    if isinstance(value, (int, float)):
-        # Possibly an epoch time.
-        return datetime.datetime.fromtimestamp(value, tzlocal())
-    else:
+    for tzinfo in get_tzinfo_options():
         try:
-            return datetime.datetime.fromtimestamp(float(value), tzlocal())
-        except (TypeError, ValueError):
-            pass
-    try:
-        # In certain cases, a timestamp marked with GMT can be parsed into a
-        # different time zone, so here we provide a context which will
-        # enforce that GMT == UTC.
-        return dateutil.parser.parse(value, tzinfos={'GMT': tzutc()})
-    except (TypeError, ValueError) as e:
-        raise ValueError('Invalid timestamp "%s": %s' % (value, e))
+            return _parse_timestamp_with_tzinfo(value, tzinfo)
+        except OSError as e:
+            logger.debug('Unable to parse timestamp with "%s" timezone info.',
+                         tzinfo.__name__, exc_info=e)
+    raise RuntimeError('Unable to calculate correct timezone offset for '
+                       '"%s"' % value)
 
 
 def parse_to_aware_datetime(value):
@@ -853,6 +918,16 @@ def is_valid_endpoint_url(endpoint_url):
     return allowed.match(hostname)
 
 
+def validate_region_name(region_name):
+    """Provided region_name must be a valid host label."""
+    if region_name is None:
+        return
+    valid_host_label = re.compile(r'^(?![0-9]+$)(?!-)[a-zA-Z0-9-]{,63}(?<!-)$')
+    valid = valid_host_label.match(region_name)
+    if not valid:
+        raise InvalidRegionError(region_name=region_name)
+
+
 def check_dns_name(bucket_name):
     """
     Check to see if the ``bucket_name`` complies with the
@@ -1105,6 +1180,12 @@ class S3RegionRedirector(object):
             # transport error.
             return
 
+        if self._is_s3_accesspoint(request_dict.get('context', {})):
+            logger.debug(
+                'S3 request was previously to an accesspoint, not redirecting.'
+            )
+            return
+
         if request_dict.get('context', {}).get('s3_redirected'):
             logger.debug(
                 'S3 request was previously redirected, not redirecting.')
@@ -1216,6 +1297,8 @@ class S3RegionRedirector(object):
         This handler retrieves a given bucket's signing context from the cache
         and adds it into the request context.
         """
+        if self._is_s3_accesspoint(context):
+            return
         bucket = params.get('Bucket')
         signing_context = self._cache.get(bucket)
         if signing_context is not None:
@@ -1223,6 +1306,289 @@ class S3RegionRedirector(object):
         else:
             context['signing'] = {'bucket': bucket}
 
+    def _is_s3_accesspoint(self, context):
+        return 's3_accesspoint' in context
+
+
+class InvalidArnException(ValueError):
+    pass
+
+
+class ArnParser(object):
+    def parse_arn(self, arn):
+        arn_parts = arn.split(':', 5)
+        if len(arn_parts) < 6:
+            raise InvalidArnException(
+                'Provided ARN: %s must be of the format: '
+                'arn:partition:service:region:account:resource' % arn
+            )
+        return {
+            'partition': arn_parts[1],
+            'service': arn_parts[2],
+            'region': arn_parts[3],
+            'account': arn_parts[4],
+            'resource': arn_parts[5],
+        }
+
+
+class S3ArnParamHandler(object):
+    _ACCESSPOINT_RESOURCE_REGEX = re.compile(
+        r'^accesspoint[/:](?P<resource_name>.+)$'
+    )
+    _BLACKLISTED_OPERATIONS = [
+        'CreateBucket'
+    ]
+
+    def __init__(self, arn_parser=None):
+        self._arn_parser = arn_parser
+        if arn_parser is None:
+            self._arn_parser = ArnParser()
+
+    def register(self, event_emitter):
+        event_emitter.register('before-parameter-build.s3', self.handle_arn)
+
+    def handle_arn(self, params, model, context, **kwargs):
+        if model.name in self._BLACKLISTED_OPERATIONS:
+            return
+        arn_details = self._get_arn_details_from_bucket_param(params)
+        if arn_details is None:
+            return
+        if arn_details['resource_type'] == 'accesspoint':
+            self._store_accesspoint(params, context, arn_details)
+
+    def _get_arn_details_from_bucket_param(self, params):
+        if 'Bucket' in params:
+            try:
+                arn = params['Bucket']
+                arn_details = self._arn_parser.parse_arn(arn)
+                self._add_resource_type_and_name(arn, arn_details)
+                return arn_details
+            except InvalidArnException:
+                pass
+        return None
+
+    def _add_resource_type_and_name(self, arn, arn_details):
+        match = self._ACCESSPOINT_RESOURCE_REGEX.match(arn_details['resource'])
+        if match:
+            arn_details['resource_type'] = 'accesspoint'
+            arn_details['resource_name'] = match.group('resource_name')
+        else:
+            raise UnsupportedS3ArnError(arn=arn)
+
+    def _store_accesspoint(self, params, context, arn_details):
+        # Ideally the access-point would be stored as a parameter in the
+        # request where the serializer would then know how to serialize it,
+        # but access-points are not modeled in S3 operations so it would fail
+        # validation. Instead, we set the access-point to the bucket parameter
+        # to have some value set when serializing the request and additional
+        # information on the context from the arn to use in forming the
+        # access-point endpoint.
+        params['Bucket'] = arn_details['resource_name']
+        context['s3_accesspoint'] = {
+            'name': arn_details['resource_name'],
+            'account': arn_details['account'],
+            'partition': arn_details['partition'],
+            'region': arn_details['region'],
+        }
+
+
+class S3EndpointSetter(object):
+    _DEFAULT_PARTITION = 'aws'
+    _DEFAULT_DNS_SUFFIX = 'amazonaws.com'
+
+    def __init__(self, endpoint_resolver, region=None,
+                 s3_config=None, endpoint_url=None, partition=None):
+        self._endpoint_resolver = endpoint_resolver
+        self._region = region
+        self._s3_config = s3_config
+        if s3_config is None:
+            self._s3_config = {}
+        self._endpoint_url = endpoint_url
+        self._partition = partition
+        if partition is None:
+            self._partition = self._DEFAULT_PARTITION
+
+    def register(self, event_emitter):
+        event_emitter.register('before-sign.s3', self.set_endpoint)
+
+    def set_endpoint(self, request, **kwargs):
+        if self._use_accesspoint_endpoint(request):
+            self._validate_accesspoint_supported(request)
+            region_name = self._resolve_region_for_accesspoint_endpoint(
+                request)
+            self._switch_to_accesspoint_endpoint(request, region_name)
+            return
+        if self._use_accelerate_endpoint:
+            switch_host_s3_accelerate(request=request, **kwargs)
+        if self._s3_addressing_handler:
+            self._s3_addressing_handler(request=request, **kwargs)
+
+    def _use_accesspoint_endpoint(self, request):
+        return 's3_accesspoint' in request.context
+
+    def _validate_accesspoint_supported(self, request):
+        if self._endpoint_url:
+            raise UnsupportedS3AccesspointConfigurationError(
+                msg=(
+                    'Client cannot use a custom "endpoint_url" when '
+                    'specifying an access-point ARN.'
+                )
+            )
+        if self._use_accelerate_endpoint:
+            raise UnsupportedS3AccesspointConfigurationError(
+                msg=(
+                    'Client does not support s3 accelerate configuration '
+                    'when and access-point ARN is specified.'
+                )
+            )
+        request_partion = request.context['s3_accesspoint']['partition']
+        if request_partion != self._partition:
+            raise UnsupportedS3AccesspointConfigurationError(
+                msg=(
+                    'Client is configured for "%s" partition, but access-point'
+                    ' ARN provided is for "%s" partition. The client and '
+                    ' access-point partition must be the same.' % (
+                        self._partition, request_partion)
+                )
+            )
+
+    def _resolve_region_for_accesspoint_endpoint(self, request):
+        if self._s3_config.get('use_arn_region', True):
+            accesspoint_region = request.context['s3_accesspoint']['region']
+            # If we are using the region from the access point,
+            # we will also want to make sure that we set it as the
+            # signing region as well
+            self._override_signing_region(request, accesspoint_region)
+            return accesspoint_region
+        return self._region
+
+    def _switch_to_accesspoint_endpoint(self, request, region_name):
+        original_components = urlsplit(request.url)
+        accesspoint_endpoint = urlunsplit((
+            original_components.scheme,
+            self._get_accesspoint_netloc(request.context, region_name),
+            self._get_accesspoint_path(
+                original_components.path, request.context),
+            original_components.query,
+            ''
+        ))
+        logger.debug(
+            'Updating URI from %s to %s' % (request.url, accesspoint_endpoint))
+        request.url = accesspoint_endpoint
+
+    def _get_accesspoint_netloc(self, request_context, region_name):
+        s3_accesspoint = request_context['s3_accesspoint']
+        accesspoint_netloc_components = [
+            '%s-%s' % (s3_accesspoint['name'], s3_accesspoint['account']),
+            's3-accesspoint'
+        ]
+        if self._s3_config.get('use_dualstack_endpoint'):
+            accesspoint_netloc_components.append('dualstack')
+        accesspoint_netloc_components.extend(
+            [
+                region_name,
+                self._get_dns_suffix(region_name)
+            ]
+        )
+        return '.'.join(accesspoint_netloc_components)
+
+    def _get_accesspoint_path(self, original_path, request_context):
+        # The Bucket parameter was substituted with the access-point name as
+        # some value was required in serializing the bucket name. Now that
+        # we are making the request directly to the access point, we will
+        # want to remove that access-point name from the path.
+        name = request_context['s3_accesspoint']['name']
+        # All S3 operations require at least a / in their path.
+        return original_path.replace('/' + name, '', 1) or '/'
+
+    def _get_dns_suffix(self, region_name):
+        resolved = self._endpoint_resolver.construct_endpoint(
+            's3', region_name)
+        dns_suffix = self._DEFAULT_DNS_SUFFIX
+        if resolved and 'dnsSuffix' in resolved:
+            dns_suffix = resolved['dnsSuffix']
+        return dns_suffix
+
+    def _override_signing_region(self, request, region_name):
+        signing_context = {
+            'region': region_name,
+        }
+        # S3SigV4Auth will use the context['signing']['region'] value to
+        # sign with if present. This is used by the Bucket redirector
+        # as well but we should be fine because the redirector is never
+        # used in combination with the accesspoint setting logic.
+        request.context['signing'] = signing_context
+
+
+    @CachedProperty
+    def _use_accelerate_endpoint(self):
+        # Enable accelerate if the configuration is set to to true or the
+        # endpoint being used matches one of the accelerate endpoints.
+
+        # Accelerate has been explicitly configured.
+        if self._s3_config.get('use_accelerate_endpoint'):
+            return True
+
+        # Accelerate mode is turned on automatically if an endpoint url is
+        # provided that matches the accelerate scheme.
+        if self._endpoint_url is None:
+            return False
+
+        # Accelerate is only valid for Amazon endpoints.
+        netloc = urlsplit(self._endpoint_url).netloc
+        if not netloc.endswith('amazonaws.com'):
+            return False
+
+        # The first part of the url should always be s3-accelerate.
+        parts = netloc.split('.')
+        if parts[0] != 's3-accelerate':
+            return False
+
+        # Url parts between 's3-accelerate' and 'amazonaws.com' which
+        # represent different url features.
+        feature_parts = parts[1:-2]
+
+        # There should be no duplicate url parts.
+        if len(feature_parts) != len(set(feature_parts)):
+            return False
+
+        # Remaining parts must all be in the whitelist.
+        return all(p in S3_ACCELERATE_WHITELIST for p in feature_parts)
+
+    @CachedProperty
+    def _addressing_style(self):
+        # Use virtual host style addressing if accelerate is enabled or if
+        # the given endpoint url is an accelerate endpoint.
+        if self._use_accelerate_endpoint:
+            return 'virtual'
+
+        # If a particular addressing style is configured, use it.
+        configured_addressing_style = self._s3_config.get('addressing_style')
+        if configured_addressing_style:
+            return configured_addressing_style
+
+    @CachedProperty
+    def _s3_addressing_handler(self):
+        # If virtual host style was configured, use it regardless of whether
+        # or not the bucket looks dns compatible.
+        if self._addressing_style == 'virtual':
+            logger.debug("Using S3 virtual host style addressing.")
+            return switch_to_virtual_host_style
+
+        # If path style is configured, no additional steps are needed. If
+        # endpoint_url was specified, don't default to virtual. We could
+        # potentially default provided endpoint urls to virtual hosted
+        # style, but for now it is avoided.
+        if self._addressing_style == 'path' or self._endpoint_url is not None:
+            logger.debug("Using S3 path style addressing.")
+            return None
+
+        logger.debug("Defaulting to S3 virtual host style addressing with "
+                     "path style addressing fallback.")
+
+        # By default, try to use virtual style with path fallback.
+        return fix_s3_host
+
 
 class ContainerMetadataFetcher(object):
 
@@ -1373,6 +1739,37 @@ def get_encoding_from_headers(headers, default='ISO-8859-1'):
         return default
 
 
+def calculate_md5(body, **kwargs):
+    if isinstance(body, (bytes, bytearray)):
+        binary_md5 = _calculate_md5_from_bytes(body)
+    else:
+        binary_md5 = _calculate_md5_from_file(body)
+    return base64.b64encode(binary_md5).decode('ascii')
+
+
+def _calculate_md5_from_bytes(body_bytes):
+    md5 = get_md5(body_bytes)
+    return md5.digest()
+
+
+def _calculate_md5_from_file(fileobj):
+    start_position = fileobj.tell()
+    md5 = get_md5()
+    for chunk in iter(lambda: fileobj.read(1024 * 1024), b''):
+        md5.update(chunk)
+    fileobj.seek(start_position)
+    return md5.digest()
+
+
+def conditionally_calculate_md5(params, **kwargs):
+    """Only add a Content-MD5 if the system supports it."""
+    headers = params['headers']
+    body = params['body']
+    if MD5_AVAILABLE and body and 'Content-MD5' not in headers:
+        md5_digest = calculate_md5(body, **kwargs)
+        params['headers']['Content-MD5'] = md5_digest
+
+
 class FileWebIdentityTokenLoader(object):
     def __init__(self, web_identity_token_path, _open=open):
         self._web_identity_token_path = web_identity_token_path
@@ -1381,3 +1778,26 @@ class FileWebIdentityTokenLoader(object):
     def __call__(self):
         with self._open(self._web_identity_token_path) as token_file:
             return token_file.read()
+
+
+class SSOTokenLoader(object):
+    def __init__(self, cache=None):
+        if cache is None:
+            cache = {}
+        self._cache = cache
+
+    def _generate_cache_key(self, start_url):
+        return hashlib.sha1(start_url.encode('utf-8')).hexdigest()
+
+    def __call__(self, start_url):
+        cache_key = self._generate_cache_key(start_url)
+        try:
+            token = self._cache[cache_key]
+            return token['accessToken']
+        except KeyError:
+            logger.debug('Failed to load SSO token:', exc_info=True)
+            error_msg = (
+                'The SSO access token has either expired or is otherwise '
+                'invalid.'
+            )
+            raise SSOTokenLoadError(error_msg=error_msg)
